/*
 *	Sherlock Gatherer Database Expiration Utility
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>

#include "gather.h"

#include <lib/url.h>
#include <lib/fastbuf.h>

static byte *log_file;
static uns refresh_after = ~0, expire_after = ~0, expire_err_after = ~0, verbose;
static uns obj_count, queue_count, requeue_count, expire_count;

static struct fastbuf *F;

static struct cfitem cfitems[] = {
  { "LogFile", CI_STRING, &log_file },
  { "RefreshAfter", CI_INT, &refresh_after },
  { "ExpireAfter", CI_INT, &expire_after },
  { "ExpireErrorsAfter", CI_INT, &expire_err_after },
  { "Verbose", CI_INT, &verbose },
  { NULL, CI_STOP, NULL }
};

static void
kill_obj(struct urlrec *r, byte *url)
{
  byte name[64];
  struct odes *d;
  struct oattr *a;
  byte *x;

  if (r->oid < OID_MIN)
    goto delurl;
  mk_obj_name("obj", r->oid, name);
  d = obj_load(name);
  a = find_attr(d, 'U');
  while (a)
    {
      if (!strcmp(a->val, url))
	break;
      a = a->same;
    }
  if (!a)
    {
      log(L_ERROR "%s: object %08x doesn't contain this URL", url, r->oid);
    erex:
      obj_free(d);
      return;
    }
  if (!del_attr(d, a))
    {
      log(L_ERROR "%s: del_attr failed", url);
      goto erex;
    }
  if (find_aval(d, 'U'))
    {
      if (verbose)
	log(L_INFO "Object %08x has more references", r->oid);
      obj_write(name, d);
      obj_free(d);
      goto delurl;
    }
  if (x = find_aval(d, 'M'))
    {
      byte md5[MD5_SIZE];
      if (verbose > 5)
	log(L_DEBUG "Deleting MD5 %s", x);
      hex_to_md5(x, md5);
      if (!md5db_delete(md5))
	{
	  log(L_ERROR "Error while deleting MD5 %s for %s", x, url);
	  goto erex;
	}
    }
  obj_free(d);
  if (unlink(name) < 0)
    log(L_ERROR "unlink(%s) failed: %m", name);

delurl:
  if (!urldb_delete(url))
    log(L_ERROR "Error while deleting URL %s", url);
}

static void
requeue_obj(struct urlrec *r, byte *url)
{
  struct url u;
  int err;
  byte b1[MAX_URL_SIZE], b2[MAX_URL_SIZE];
  struct qhost *h;

  err = url_canon_split(url, b1, b2, &u);
  if (err)
    {
      log(L_ERROR "%s: %s", url, url_error(err));
      return;
    }
  h = find_host(u.protoid, u.host, u.port);
  if (!h)
    {
      log(L_ERROR "%s: Host %d:%s:%d not found", url, u.protoid, u.host, u.port);
      return;
    }
  enqueue_item(h, u.rest);
  enqueue_host(h);
  urldb_store(url, r->access|1, r->oid);
}

static void
expire_fill(void)
{
  byte buf[MAX_URL_SIZE];
  struct urlrec *r;
  time_t now = time(NULL);
  uns age;

  if (!urldb_get_first(buf))
    return;
  do
    {
      if (!(r = urldb_lookup(buf)))
	die("%s: No data found", buf);
      obj_count++;
      if (r->access & 1)
	{
	  if (verbose > 5)
	    log(L_DEBUG "%s: queued", buf);
	  queue_count++;
	  continue;
	}
      if (r->access > now)
	{
	  log(L_ERROR "%s: Ante-dated (access=%d, now=%d)", buf, r->access, now);
	  continue;
	}
      age = now - r->access;
      if (r->oid < OID_MIN)		/* It's an error */
	{
	  if (age >= expire_err_after)
	    {
	      if (verbose)
		log(L_INFO "%s [%d]: Expiring (age=%d)", buf, r->oid, age);
	      expire_count++;
	      bputc(F, '-');
	      bputsn(F, buf);
	    }
	  else
	    {
	      if (verbose > 5)
		log(L_DEBUG "%s [%d]: OK", r->oid, buf);
	    }
	}
      else if (age >= expire_after)
	{
	  if (verbose)
	    log(L_INFO "%s: Expiring (age=%d)", buf, age);
	  expire_count++;
	  bputc(F, '-');
	  bputsn(F, buf);
	}
      else if (age >= refresh_after)
	{
	  if (verbose)
	    log(L_INFO "%s: Requeued (age=%d)", buf, age);
	  requeue_count++;
	  bputc(F, '+');
	  bputsn(F, buf);
	}
      else
	{
	  if (verbose > 5)
	    log(L_DEBUG "%s: OK", buf);
	}
    }
  while (urldb_get_next(buf));
}

static void
expire_read(void)
{
  byte buf[MAX_URL_SIZE+4];
  struct urlrec *r;
  byte *url = buf+1;

  while (bgets(F, buf, sizeof(buf)))
    {
      if (verbose > 3)
	puts(buf);
      if (!(r = urldb_lookup(url)))
	die("%s: No data found", url);
      switch (buf[0])
	{
	case '+':
	  requeue_obj(r, url);
	  break;
	case '-':
	  kill_obj(r, url);
	  break;
	default:
	  die("%s: Unknown operation");
	}
    }
}

int
main(int argc, char **argv)
{
  initlog(argv[0]);
  if (argc != 1)
    die("Whaddya want to do? Gimme a config file and that's all!");
  cf_read("cf/expire", cfitems);
  open_log_file(log_file);
  log(L_INFO "Expire started");

  if (!(F = bopen("tmp/expire.tmp", O_CREAT | O_TRUNC | O_RDWR, 65536)))
    die("Cannot open tmp file: %m");

  gather_lock();
  queue_init();
  urldb_open();
  md5db_open();

  expire_fill();
  bflush(F);
  bsetpos(F, 0);
  log(L_INFO "Pass 2");
  expire_read();
  bclose(F);
  unlink("tmp/expire.tmp");

  queue_cleanup();
  urldb_close();
  md5db_close();
  gather_unlock();
  log(L_INFO "Expire done: seen %d objects, %d already queued, requeued %d, expired %d.",
      obj_count, queue_count, requeue_count, expire_count);
  return 0;
}

