/*
 *	Sherlock Object Distributor -- The Give Daemon
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <syslog.h>
#include <signal.h>
#include <sys/stat.h>
#include <dirent.h>

#include <lib/lib.h>
#include <lib/string.h>

static byte peer_name[128];
static int authentic = 0;

static uns init_timeout = 240;
static uns read_timeout = 120;
static uns write_timeout = 240;

static byte *
cf_access(struct cfitem *c, byte *p)
{
  if (!authentic && match_pattern_nocase(p, peer_name))
    {
      authentic = 1;
      if (c->name[0] == 'D')
	{
	  puts("-Access available, but not for you");
	  syslog(LOG_ERR, "Access denied");
	  exit(1);
	}
    }
  return NULL;
}

static struct cfitem cfitems[] = {
  { "Allow", CI_FUNCTION, cf_access },
  { "Deny", CI_FUNCTION, cf_access },
  { "InitTimeout", CI_INT, &init_timeout },
  { "ReadTimeout", CI_INT, &read_timeout },
  { "WriteTimeout", CI_INT, &write_timeout },
  { NULL, CI_STOP, NULL }
};

static void
timeout(int unused)
{
  syslog(LOG_ERR, "Session timed out");
  exit(1);
}

static void
starttimer(int i)
{
  signal(SIGALRM, timeout);
  alarm(i);
}

static void
transfer(uns from)
{
  uns total_objs = 0;
  uns total_bytes = 0;
  DIR *d, *g;
  struct dirent *dd, *gg;
  byte dirname[16], objname[16];
  struct stat st;
  int fd, i;
  byte buf[4096];

  alarm(0);
  if (! (d = opendir("obj")))
    {
      puts("-Local configuration error");
      syslog(LOG_ERR, "obj directory not found");
      exit(1);
    }
  while (dd = readdir(d))
    {
      if (!Cxdigit(dd->d_name[0]) || !Cxdigit(dd->d_name[1]) || dd->d_name[2])
	continue;
      sprintf(dirname, "obj/%s", dd->d_name);
      if (! (g = opendir(dirname)))
	{
	  syslog(LOG_ERR, "%s scanned, but not found", dirname);
	  continue;
	}
      while (gg = readdir(g))
	{
	  if (strlen(gg->d_name) != 6 || !Cxdigit(gg->d_name[0]) || !Cxdigit(gg->d_name[1])) /* Should be enough */
	    continue;
	  sprintf(objname, "%s/%s", dirname, gg->d_name);
	  if (stat(objname, &st))
	    {
	      syslog(LOG_ERR, "%s: stat failed: %m", objname);
	      continue;
	    }
	  if (! (st.st_mode & S_IROTH))	/* Currently being written to */
	    continue;
	  if (st.st_mtime < from)	/* Too old */
	    continue;
	  if ((fd = open(objname, O_RDONLY, 0)) < 0)
	    {
	      syslog(LOG_ERR, "%s: open failed: %m", objname);
	      continue;
	    }
	  starttimer(write_timeout);
	  printf("+%s%s %u\n", dd->d_name, gg->d_name, (uns) st.st_mtime);
	  while ((i = read(fd, buf, sizeof(buf))) > 0)
	    if (fwrite(buf, 1, i, stdout) != i)
	      {
		syslog(LOG_ERR, "%s: write error: %m", objname);
		exit(1);
	      }
	  close(fd);
	  putchar('\n');
	  putchar('\n');
	  total_objs++;
	  total_bytes += st.st_size;
	  alarm(0);
	}
      closedir(g);
    }
  closedir(d);
  syslog(LOG_ERR, "Transferred %d objects (%d bytes)", total_objs, total_bytes);
  printf("+ That's all: %d objects, %d bytes\n", total_objs, total_bytes);
}

int
main(int argc, char **argv)
{
  struct sockaddr sa;
  int nal = sizeof(sa);
  struct hostent *he;
  struct in_addr *ia;
  uns req;

  starttimer(init_timeout);
  openlog(argv[0], LOG_NDELAY | LOG_PID, LOG_DAEMON);
  if (argc != 1 && argc != 2)
    {
      syslog(LOG_ERR, "Invalid arguments. Expecting: gived [<directory>]");
      return 1;
    }
  if (argc == 2)
    if (chdir(argv[1]) < 0)
      {
	syslog(LOG_ERR, "chdir(%s) failed: %m", argv[1]);
	return 1;
      }

  if (getpeername(0, &sa, &nal) < 0 || sa.sa_family == AF_UNIX)
    strcpy(peer_name, "localhost");
  else if (sa.sa_family != AF_INET)
    {
      puts("-I refuse to talk anything else than TCP");
      syslog(LOG_ERR, "Something else than TCP?");
      return 1;
    }
  else if (he = gethostbyaddr((char *) (ia = &((struct sockaddr_in *)&sa)->sin_addr), sizeof(struct in_addr), AF_INET))
    {
      char **z;
      if (strlen(he->h_name) > sizeof(peer_name) - 1)
	{
	  puts("-Your name is too long for me!");
	  syslog(LOG_ERR, "Peer name too long");
	  return 1;
	}
      strcpy(peer_name, he->h_name);
      he = gethostbyname(he->h_name);
      if (!he)
	{
	  printf("-Your name %s doesn't resolve back...\n", he->h_name);
	  syslog(LOG_ERR, "No back-reverse mapping for %08x", ia->s_addr);
	  return 1;
	}
      for(z = he->h_addr_list; *z; z++)
	if (!memcmp(*z, &ia->s_addr, sizeof(struct in_addr)))
	  break;
      if (!*z)
	{
	  printf("-Your name %s resolves to different address. Bye!\n", peer_name);
	  syslog(LOG_ERR, "Fake reverse mapping for %08x", ia->s_addr);
	  return 1;
	}
    }
  else
    {
      puts("-Unable to identify you");
      syslog(LOG_ERR, "No reverse mapping for %08x", ia->s_addr);
      return 1;
    }

  syslog(LOG_INFO, "Connected to %s", peer_name);
  cf_read("cf/gived", cfitems);
  printf("+Hello, %s. Pleased to meet you. Tell me your magic number...\n", peer_name);
  fflush(stdout);

  starttimer(read_timeout);
  if (scanf("%d", &req) != 1)
    {
      puts("-Black magic!");
      syslog(LOG_ERR, "No request.");
      return 1;
    }
  transfer(req);
  return 0;
}
