/*
 *	Sherlock Library -- Object Functions
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "lib.h"
#include "pools.h"

#define OBJ_POOL_SIZE 4096

void
obj_dump(struct odes *o)
{
  struct oattr *a, *b;

  for(a=o->attrs; a; a=a->next)
    for(b=a; b; b=b->same)
      printf("%c%s\n", (a==b ? a->attr : ' '), b->val);
}

static struct oattr *
oa_new(struct odes *o, uns x, byte *v)
{
  struct oattr *a = pool_alloc(o->pool, sizeof(struct oattr) + strlen(v));

  a->next = a->same = NULL;
  a->attr = x;
  strcpy(a->val, v);
  return a;
}

struct odes *
obj_new(void)
{
  struct mempool *p = new_pool(OBJ_POOL_SIZE);
  struct odes *o = pool_alloc(p, sizeof(struct odes));

  o->pool = p;
  o->attrs = NULL;
  return o;
}

struct odes *
obj_fload(FILE *f)
{
  byte buf[1024];
  byte *k;
  struct odes *o = obj_new();
  struct oattr **last = &o->attrs;
  struct oattr *a, *la;

  la = NULL;
  *last = NULL;
  while (fgets(buf, sizeof(buf), f))
    {
      k = strchr(buf, '\n');
      if (!k)
	die("Malformed object, line too long?");
      *k = 0;
      if (!buf[0])
	die("Malformed object: empty line");
      a = oa_new(o, buf[0], buf+1);
      if (!la || la->attr != a->attr)
	for(la=o->attrs; la && la->attr != a->attr; la=la->next)
	  ;
      if (la)
	{
	  while (la->same)
	    la = la->same;
	  la->same = a;
	}
      else
	{
	  *last = a;
	  last = &a->next;
	}
      la = a;
    }
  return o;
}

struct odes *
obj_load(byte *name)
{
  FILE *f;
  struct odes *d;

  if (! (f = fopen(name, "r")))
    die("Unable to open object %s: %m", name);
  d = obj_fload(f);
  fclose(f);
  return d;
}

void
obj_fwrite(FILE *f, struct odes *d)
{
  struct oattr *a, *b;
  byte *z;

  for(a=d->attrs; a; a=a->next)
    for(b=a; b; b=b->same)
      {
	fputc(a->attr, f);
	for(z = b->val; *z; z++)
	  if (*z >= ' ')
	    fputc(*z, f);
	  else
	    {
	      fputc('?', f);
	      log(L_ERROR "obj_dump: Found non-ASCII characters (URL might be %s)", find_aval(d, 'U'));
	    }
	fputc('\n', f);
      }
}

void
obj_write(byte *name, struct odes *d)
{
  FILE *f;

  if (! (f = fopen(name, "w")))
    die("Unable to write object %s: %m", name);
  obj_fwrite(f, d);
  fclose(f);
}

void
obj_free(struct odes *o)
{
  free_pool(o->pool);
}

struct oattr *
find_attr(struct odes *o, uns x)
{
  struct oattr *a;

  for(a=o->attrs; a && a->attr != x; a=a->next)
    ;
  return a;
}

struct oattr *
find_attr_last(struct odes *o, uns x)
{
  struct oattr *a = find_attr(o, x);

  if (a)
    {
      while (a->same)
	a = a->same;
    }
  return a;
}

uns
del_attr(struct odes *o, struct oattr *a)
{
  struct oattr *x, **p;
  byte aa = a->attr;

  p = &o->attrs;
  while (x = *p)
    {
      if (x->attr == aa)
	{
	  while (x = *p)
	    {
	      if (x == a)
		{
		  *p = x->same;
		  return 1;
		}
	      p = &x->same;
	    }
	  return 0;
	}
      p = &x->next;
    }
  return 0;
}

byte *
find_aval(struct odes *o, uns x)
{
  struct oattr *a = find_attr(o, x);

  if (a)
    return a->val;
  else
    return NULL;
}

struct oattr *
set_attr(struct odes *o, uns x, byte *v)
{
  struct oattr *a, **z;

  z = &o->attrs;
  while (a = *z)
    {
      if (a->attr == x)
	{
	  *z = a->next;
	  /* free_achain(a); */
	  goto set;
	}
      z = &a->next;
    }

 set:
  if (v)
    {
      a = oa_new(o, x, v);
      a->next = o->attrs;
      o->attrs = a;
    }
  else
    a = NULL;
  return a;
}

struct oattr *
set_attr_num(struct odes *o, uns a, uns v)
{
  byte x[32];

  sprintf(x, "%d", v);
  return set_attr(o, a, x);
}

struct oattr *
add_attr(struct odes *o, struct oattr *a, uns x, byte *v)
{
  struct oattr *b;

  if (!a)
    {
      a = find_attr(o, x);
      if (!a)
	return set_attr(o, x, v);
    }
  while (a->same)
    a = a->same;
  b = oa_new(o, x, v);
  a->same = b;
  return b;
}

struct oattr *
prepend_attr(struct odes *o, uns x, byte *v)
{
  struct oattr *b, *a, **z;

  b = oa_new(o, x, v);
  z = &o->attrs;
  while (a = *z)
    {
      if (a->attr == x)
	{
	  b->next = a->next;
	  a->next = NULL;
	  b->same = a;
	  *z = b;
	  return b;
	}
      z = &a->next;
    }
  b->next = o->attrs;
  o->attrs = b;
  return b;
}
