/*
 *	Sherlock Database Builder -- Hashing of Non-Words
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>

#include <lib/lib.h>
#include <lib/searchdb.h>
#include <lib/fastbuf.h>

#define BUFSIZE 65536

#undef NHASH_DEBUG

#ifdef NHASH_DEBUG
#define DBG(x, args...) log(L_DEBUG x, ## args)
#else
#define DBG(x, args...)
#endif

static struct fastbuf *fi, *fo, *idx;
static ulg count, size;

/* Configuration */

static uns magnification = 10;
static uns num_cache_words = 1024;
static uns cache_buf_size = 65536;

static struct cfitem cfitems[] = {
  { "Magnification", CI_INT, &magnification },
  { "CacheWords", CI_INT, &num_cache_words },
  { "CacheBuffer", CI_INT, &cache_buf_size },
  { NULL, CI_STOP, NULL }
};

/* Calculating hash table size... */

static void
sizing(void)
{
  bseek(fi, -sizeof(count), SEEK_END);
  count = bgetl_slow(fi);
  log(L_INFO "Building non-word hash table, %d items", count);
  size = nextprime(count * magnification / 10);
  log(L_INFO "Hash table size set to %d", size);
  bsetpos(fi, 0);
}

/* MERGESORT (shamelessly stolen from sword.c) */

struct fib {
  struct fastbuf *f;
  byte lastword[MAX_WORD_LEN];
  byte name[32];
  int eof;
  ulg hash;
};

static struct fib files[4];
static struct fib *src[2], *dest[2];

static void
opendest(void)
{
  int i;

  for(i=0; i<2; i++)
    {
      struct fib *x = dest[i];
      int retry = 0;
      int fd;
      do
	{
	  if (retry++ > 50)
	    die("Unable to create temporary file, tried %d times (%m)", retry);
	  sprintf(x->name, "nhash-%d.%08x", getpid(), temprand(retry));
	  fd = open(x->name, O_RDWR | O_EXCL | O_CREAT, 0666);
	}
      while (fd < 0);
      x->f = bfdopen(fd, BUFSIZE);
      x->lastword[0] = x->eof = x->hash = 0;
      DBG("Created output file `%s', channel #%d", x->name, i);
    }
}

static inline void
fill_in(struct fib *r)
{
  byte *p;
  int i;
  uns h = 0;
  struct fastbuf *f = r->f;

  if (r->eof || r->lastword[0])
    return;
  p = r->lastword;
  do
    {
      if (f->bptr < f->bstop)
	i = *f->bptr++;
      else if ((i = bgetc_slow(f)) == EOF)
	{
	  r->eof = 1;
	  r->lastword[0] = 0;
	  return;
	}
      *p++ = i;
      h = (h * HASH_MAGIC + i) % size;
    }
  while (i);
  r->hash = h;
}

static void
merge(void)
{
  struct fib *r1, *r2, *r, *w1, *w2;

  r1 = src[0];
  r2 = src[1];
  r1->lastword[0] = r2->lastword[0] = 0;
  if (!r2->f)
    r2->eof = 1;
  w1 = dest[0];
  w2 = dest[1];
  for(;;)
    {
      fill_in(r1);
      fill_in(r2);
      if (!r1->eof)			/* Select input channel... */
	{
	  if (!r2->eof)
	    {				/* Both channels non-empty */
	      if (r1->hash <= r2->hash)
		r = r1;
	      else
		r = r2;
	    }
	  else
	    r = r1;
	}
      else if (!r2->eof)
	r = r2;
      else				/* The world is coming to an end... */
	break;
      if (w1->hash > r->hash)		/* Need to swap files */
	{
	  struct fib *w;
	  w = w1; w1 = w2; w2 = w;
	}
      w1->hash = r->hash;
      w1->lastword[0] = 1;		/* MAGIC (mark output channel as used) */
      bputs(w1->f, r->lastword);
      bputc(w1->f, 0);
      bputl(w1->f, bgetl(r->f));
      r->lastword[0] = 0;
    }
  bflush(w1->f);
  bflush(w2->f);
}

static void
closesrc(void)
{
  int i;

  for(i=0; i<2; i++)
    {
      struct fib *x = src[i];

      if (x->f)
	{
	  DBG("Closing `%s'", x->name);
	  bclose(x->f);
	  if (x->name[0] && unlink(x->name))
	    log(L_ERROR "unlink(`%s'): %m", x->name);
	}
    }
}

static void
swapsrcdest(void)
{
  struct fib *a;

  DBG("Swapping...");
  a = src[0]; src[0] = dest[0]; dest[0] = a;
  a = src[1]; src[1] = dest[1]; dest[1] = a;
  bsetpos(src[0]->f, 0);
  bsetpos(src[1]->f, 0);
}

static void
closefinal(void)
{
  struct fib *x;
  ulg lasth = ~0;
  ulg ref;
  uns coll = 0;
  uns total = 0;
  uns maxcoll = 0;

  swapsrcdest();
  DBG("Final output...");
  x = src[0];
  x->lastword[0] = 0;
  for(;;)
    {
      fill_in(x);
      if (x->eof)
	break;
      ref = bgetl(x->f);
      if (x->hash != lasth)
	{
	  if (lasth != ~0)
	    bputc(fo, 0);
	  while (++lasth < x->hash)
	    bputl(idx, ~0);
	  bputl(idx, btell(fo));
	  total++;
	  coll = 0;
	}
      else if (++coll > maxcoll)
	maxcoll = coll;
      bputs(fo, x->lastword);
      bputc(fo, 0);
      bputl(fo, ref);
      x->lastword[0] = 0;
    }
  bputc(fo, 0);
  while (++lasth < size)
    bputl(idx, ~0);
  log(L_INFO "Hash table built. Used %d slots, got maximal chain of %d items.", total, maxcoll+1);
  closesrc();				/* Delete everything */
}

/* Pre-sorting cache */

struct cache_item {
  ulg ref;
  ulg hash;
  byte word[1];
};

static byte *cache_buffer, *cache_buf_end, *cache_free;
static struct cache_item **cache_index;

static int
ci_cmp(struct cache_item **a, struct cache_item **b)
{
  struct cache_item *A = *a;
  struct cache_item *B = *b;

  if (A->hash < B->hash)
    return -1;
  if (A->hash > B->hash)
    return 1;
  return 0;
}

static void
presort(void)
{
  struct fib *r, *w;
  uns wi, wcnt, l;
  struct cache_item *i;

  cache_free = cache_buffer = xmalloc(cache_buf_size);
  cache_buf_end = cache_buffer + cache_buf_size;
  cache_index = xmalloc(num_cache_words * sizeof(struct cache_item *));
  wcnt = wi = 0;

  r = src[0];
  while (!r->eof)
    {
      w = dest[wi];

      for(;;)
	{
	  fill_in(r);
	  if (r->eof)
	    break;
	  if (wcnt >= num_cache_words)
	    break;
	  l = sizeof(struct cache_item) + strlen(r->lastword);
	  l = (l + CPU_STRUCT_ALIGN - 1) & ~(CPU_STRUCT_ALIGN - 1);
	  if (cache_free + l > cache_buf_end)
	    break;
	  i = (struct cache_item *) cache_free;
	  cache_free += l;
	  i->ref = bgetl(r->f);
	  i->hash = r->hash;
	  strcpy(i->word, r->lastword);
	  cache_index[wcnt++] = i;
	  r->lastword[0] = 0;
	}

      if (wcnt > 0)
	{
	  qsort(cache_index, wcnt, sizeof(struct cache_item *), (int (*)(const void *, const void *)) ci_cmp);
	  for(l=0; l<wcnt; l++)
	    {
	      i = cache_index[l];
	      bputs(w->f, i->word);
	      bputc(w->f, 0);
	      bputl(w->f, i->ref);
	    }
	  cache_free = cache_buffer;
	  wcnt = 0;
	}

      w->lastword[0] = 1;		/* MAGIC */
      wi = !wi;
    }

  bflush(dest[0]->f);
  bflush(dest[1]->f);
  free(cache_buffer);
  free(cache_index);
}

int
main(int argc, char **argv)
{
  int pass = 0;

  initlog(argv[0]);
  if (argc != 4)
    die("Usage: nhash <nonword-list> <hash-list> <hash-index>");
  fi = bopen(argv[1], O_RDONLY, BUFSIZE);
  fo = bopen(argv[2], O_WRONLY | O_CREAT | O_TRUNC, BUFSIZE);
  idx = bopen(argv[3], O_WRONLY | O_CREAT | O_TRUNC, BUFSIZE);
  cf_read("cf/nhash", cfitems);
  sizing();
  src[0] = files;
  src[1] = files + 1;
  dest[0] = files + 2;
  dest[1] = files + 3;
  src[0]->f = fi;
  for(;;)
    {
      opendest();
      if (pass++)
	merge();
      else
	presort();
      closesrc();
      if (!dest[1]->lastword[0])
	break;
      swapsrcdest();
    }
  closefinal();
  bclose(fo);
  bclose(idx);
  return 0;
}
