/*
 *	Sherlock Database Builder -- Word Reference File Sorting
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <lib/string.h>
#include <lib/lib.h>
#include <lib/searchdb.h>
#include <lib/wordcmp.h>
#include <lib/fastbuf.h>

#undef SWORD_DEBUG

#ifdef SWORD_DEBUG
#define DBG(x, args...) log(L_DEBUG x, ## args)
#else
#define DBG(x, args...)
#endif

#define BUFSIZE 65536

struct fib {
  struct fastbuf *f;
  byte lastword[MAX_WORD_LEN];
  byte name[32];
  int eof;
};

static struct fib files[4];
static struct fib *src[2], *dest[2];
static uns nonword_flag;

static inline int
compare(byte *a, byte *b)
{
  return nonword_flag ? strcmp(a, b) : search_wordcmp(a, b);
}

static void
opendest(void)
{
  int i;

  for(i=0; i<2; i++)
    {
      struct fib *x = dest[i];
      int retry = 0;
      int fd;
      do
	{
	  if (retry++ > 50)
	    die("Unable to create temporary file, tried %d times (%m)", retry);
	  sprintf(x->name, "sword-%d.%08x", getpid(), temprand(retry));
	  fd = open(x->name, O_RDWR | O_EXCL | O_CREAT, 0666);
	}
      while (fd < 0);
      x->f = bfdopen(fd, BUFSIZE);
      x->lastword[0] = x->eof = 0;
      DBG("Created output file `%s', channel #%d", x->name, i);
    }
}

static inline void
fill_in(struct fib *r)
{
  byte *p;
  int i;
  struct fastbuf *f = r->f;

  if (r->eof || r->lastword[0])
    return;
  p = r->lastword;
  do
    {
      if (f->bptr < f->bstop)
	i = *f->bptr++;
      else if ((i = bgetc_slow(f)) == EOF)
	{
	  r->eof = 1;
	  r->lastword[0] = 0;
	  return;
	}
      *p++ = i;
    }
  while (i);
}

static void
merge(void)
{
  struct fib *r1, *r2, *r, *w1, *w2;
  int flag;

  r1 = src[0];
  r2 = src[1];
  r1->lastword[0] = r2->lastword[0] = 0;
  if (!r2->f)
    r2->eof = 1;
  w1 = dest[0];
  w2 = dest[1];
  for(;;)
    {
      fill_in(r1);
      fill_in(r2);
      flag = 0;
      if (!r1->eof)			/* Select input channel... */
	{
	  if (!r2->eof)
	    {				/* Both channels non-empty */
	      int c = compare(r1->lastword, r2->lastword);
	      if (c < 0)
		r = r1;
	      else if (c > 0)
		r = r2;
	      else			/* Need to merge! */
		{
		  r = r1;
		  flag = 1;
		}
	    }
	  else
	    r = r1;
	}
      else if (!r2->eof)
	r = r2;
      else				/* The world is coming to an end... */
	break;
      if (compare(w1->lastword, r->lastword) >= 0) /* Need to swap files */
	{
	  struct fib *w;
	  w = w1; w1 = w2; w2 = w;
	}
      strcpy(w1->lastword, r->lastword);
      bputs(w1->f, r->lastword);
      bputc(w1->f, 0);
      if (!flag)			/* Copy only */
	{
	  ulg l;
	  l = bgetl(r->f);
	  bputl(w1->f, l);
	  bbcopy(r->f, w1->f, l * sizeof(struct sdbitem));
	  r->lastword[0] = 0;
	}
      else				/* Merge */
	{
	  ulg l1, l2, l;
	  struct sdbitem i1, i2;
	  l1 = bgetl(r1->f);
	  l2 = bgetl(r2->f);
	  if (!l1 || !l2)
	    die("Internal error: zero-length reference list spotted!");
	  l = l1 + l2;
	  bputl(w1->f, l);
	  bread(r1->f, &i1, sizeof(i1));
	  bread(r2->f, &i2, sizeof(i2));
	  while (l1 && l2)
	    {
	      if (i1.oid < i2.oid
		  || i1.oid == i2.oid && i1.flags < i2.flags)
		{			/* If OID's match, tags must match, too (!!) */
		  bwrite(w1->f, &i1, sizeof(i1));
		  if (--l1)
		    bread(r1->f, &i1, sizeof(i1));
		}
	      else
		{
		  bwrite(w1->f, &i2, sizeof(i2));
		  if (--l2)
		    bread(r2->f, &i2, sizeof(i2));
		}
	    }
	  if (l1)
	    {
	      bwrite(w1->f, &i1, sizeof(i1));
	      bbcopy(r1->f, w1->f, (l1-1) * sizeof(struct sdbitem));
	    }
	  else if (l2)
	    {
	      bwrite(w1->f, &i2, sizeof(i2));
	      bbcopy(r2->f, w1->f, (l2-1) * sizeof(struct sdbitem));
	    }
	  r1->lastword[0] = r2->lastword[0] = 0;
	}
    }
  bflush(w1->f);
  bflush(w2->f);
}

static void
closesrc(void)
{
  int i;

  for(i=0; i<2; i++)
    {
      struct fib *x = src[i];

      if (x->f)
	{
	  DBG("Closing `%s'", x->name);
	  bclose(x->f);
	  if (x->name[0] && unlink(x->name))
	    log(L_ERROR "unlink(`%s'): %m", x->name);
	}
    }
}

static void
swapsrcdest(void)
{
  struct fib *a;

  DBG("Swapping...");
  a = src[0]; src[0] = dest[0]; dest[0] = a;
  a = src[1]; src[1] = dest[1]; dest[1] = a;
  bsetpos(src[0]->f, 0);
  bsetpos(src[1]->f, 0);
}

static void
closefinal(byte *name)
{
  struct fib *x = dest[0];

  bclose(x->f);
  DBG("Renaming `%s' to dest file `%s'", x->name, name);
  if (rename(x->name, name))
    die("rename(`%s', `%s'): %m", x->name, name);

  x = dest[1];
  DBG("And unlinking dest `%s'", x->name);
  bclose(x->f);
  if (unlink(x->name))
    die("unlink(`%s'): %m", x->name);
}

int
main(int argc, char **argv)
{
  initlog(argv[0]);
  if (argc > 1 && !strcmp(argv[1], "-n"))
    {
      nonword_flag = 1;
      argc--;
      argv++;
    }
  if (argc < 3)
    die("Usage: sword [-n] <src> <dest>");
  log(L_INFO "Sorting %sword references", nonword_flag ? "non-" : "");
  src[0] = files;
  src[1] = files + 1;
  dest[0] = files + 2;
  dest[1] = files + 3;
  src[0]->f = bopen(argv[1], O_RDONLY, BUFSIZE);
  for(;;)
    {
      opendest();
      merge();
      closesrc();
      if (!dest[1]->lastword[0])
	break;
      swapsrcdest();
      log(L_DEBUG "Pass...");
    }
  closefinal(argv[2]);
  return 0;
}
