/*
 *	Sherlock Search Engine -- Query Processing
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#include "sherlock.h"

#include <lib/searchdb.h>
#include <lib/pools.h>

/* Generation of normalized query strings for easy cache lookup and debugging output */

static byte *
norm_expr(byte *b, struct expr *x)
{
  switch (x->type)
    {
    case EX_AND:
    case EX_OR:
      *b++ = '(';
      b = norm_expr(b, x->u.op.l);
      *b++ = (x->type == EX_AND) ? '&' : '|';
      b = norm_expr(b, x->u.op.r);
      *b++ = ')';
      break;
    case EX_NOT:
      *b++ = '!';
      b = norm_expr(b, x->u.op.l);
      break;
    case EX_MATCH:
      b += sprintf(b, "%d/%02x/%d:\"%s\"", x->u.match.quality,
		   x->u.match.classmap, x->u.match.is_nonword, x->u.match.word);
      break;
    case EX_ANY:
      b += sprintf(b, "ANY");
      break;
    default:
      die("Confused by gazing at unknown type %d", x->type);
    }
  return b;
}

void
normalize_query(struct query *q, byte *buf)
{
  int i;

  for(i=0; i<256/8; i++)
    buf += sprintf(buf, "%02x", q->sections[i]);
  *buf++ = ' ';
  *norm_expr(buf, q->expr) = 0;
}

/* Query cache */

struct cache_item {
  node h;				/* In hash queue (_must_ be first!) */
  node n;				/* In LRU queue */
  byte *request;			/* Normalized request */
  ulg *oids;				/* Object ID's of resulting objects */
  ulg *qs;				/* Quality factors of resulting objects */
  uns len;				/* Number of objects found */
  time_t timestamp;			/* Timestamp of last successful fill */
  struct mempool *pool;			/* Pool for cached replies */
  struct replybuf rbuf;			/* Cached replies */
};

static uns hash_count, ulg_bits_log;
static list cache_lru, cache_free, *cache_hash;

static int process_query(struct query *, struct cache_item *);
static void show_objects(struct query *, struct cache_item *, uns, uns);

void
cache_init(void)
{
  uns i;

  hash_count = nextprime(cache_size);
  ulg_bits_log = log2(8 * sizeof(ulg));
  DBG("Ulg has %d bits, log=%d", 8*sizeof(ulg), ulg_bits_log);
  log(L_INFO "Allocating cache for %d items hashed to %d boxes", cache_size, hash_count);
  init_list(&cache_lru);
  init_list(&cache_free);
  cache_hash = xmalloc(sizeof(list) * hash_count);
  for(i=0; i<hash_count; i++)
    init_list(&cache_hash[i]);
  for(i=0; i<cache_size; i++)
    {
      struct cache_item *a = xmalloc(sizeof(struct cache_item));
      a->oids = xmalloc(sizeof(ulg) * num_matches);
      a->qs = xmalloc(sizeof(ulg) * num_matches);
      add_tail(&cache_free, &a->n);
    }
}

static uns
get_hash_val(byte *x)
{
  uns a = strlen(x);

  while (*x)
    a = 23*a + *x++;
  return a % hash_count;
}

static struct cache_item *
get_free_item(void)
{
  struct cache_item *c = SKIP_BACK(struct cache_item, n, HEAD(cache_free)); /* Ugly trick */

  if (c->n.next)
    rem_node(&c->n);
  else
    {
      c = SKIP_BACK(struct cache_item, n, HEAD(cache_lru)); /* :-( */
      if (!c->n.next)
	die("Out of cache slots. Ouch!");
      rem_node(&c->n);
      rem_node(&c->h);
      free_pool(c->pool);
    }
  return c;
}

static struct cache_item *
get_cache_item(struct query *q, time_t now)
{
  byte buf[8192];
  struct cache_item *c;
  uns hash;

  normalize_query(q, buf);
  if (log_requests > 9)
    log(L_DEBUG ": %s", buf);
  hash = get_hash_val(buf);
  DO_FOR_ALL(c, cache_hash[hash])
    if (!strcmp(c->request, buf) && (now - c->timestamp) < max_cache_age)
      {
	rem_node(&c->n);
	add_tail(&cache_lru, &c->n);
	return c;
      }
  c = get_free_item();
  c->timestamp = 0;
  c->pool = new_pool(1024);
  c->request = pool_alloc(c->pool, strlen(buf) + 1);
  strcpy(c->request, buf);
  add_tail(&cache_lru, &c->n);
  add_tail(&cache_hash[hash], &c->h);
  init_reply_buf(&c->rbuf, c->pool);
  return c;
}

void
do_query(struct query *q)
{
  time_t now = time(NULL);
  struct cache_item *c = get_cache_item(q, now);
  uns start, count;

  q->parse_msec = get_timer();
  cache_rbuf = &c->rbuf;
  if (!c->timestamp)
    {
      if (!process_query(q, c))
	return;				/* Error */
      c->timestamp = now;
      add_reply(cache_rbuf, "S%d", c->len);
      q->refs_msec = get_timer();
    }
  else
    add_reply(query_rbuf, "C%d", now - c->timestamp);

  prepend_reply(query_rbuf, "+OK");
  start = q->from;
  if (start > c->len)
    start = c->len;
  count = q->to;
  if (count < start)
    count = start;
  if (count >= c->len)
    count = c->len - 1;
  count = count - start + 1;
  add_reply(query_rbuf, "N%d", count);
  flush_reply_buf(q, query_rbuf);
  ship_reply_buf(q, cache_rbuf);

  reply_string(q, "\n", 1);
  show_objects(q, c, start, count);
  q->objs_msec = get_timer();

  add_reply(query_rbuf, "t%d\t%d\t%d\t%d\t%d",
	    q->parse_msec, q->map_msec, q->words_msec, q->refs_msec, q->objs_msec);
  if (log_timing)
    log(L_INFO "Timing: parse=%d map=%d words=%d refs=%d[%d] objs=%d[%d] total=%d",
	q->parse_msec, q->map_msec, q->words_msec, q->refs_msec, c->len, q->objs_msec, count,
	q->parse_msec + q->map_msec + q->words_msec + q->refs_msec + q->objs_msec);
}

/* List of known words */

struct known_word {
  byte text[MAX_WORD_LEN];
  uns is_nonword;
  ulg refs[HARD_MAX_WORD_MATCHES];	/* Pointers to the reference index file */
  int nrefs;
  struct known_word_ref *firstr;
};

struct known_word_ref {
  struct known_word_ref *next;
  struct known_word *kw;
  ulg classmap;
  uns qfactor;
  uns qtemp;
  ulg mask;
};

static struct known_word words[HARD_MAX_WORDS];
static struct known_word_ref wordrefs[HARD_MAX_WORDS];
static uns num_words, num_word_refs;

static int
find_words(struct expr *x)
{
  uns i;
  struct known_word *w;

  switch (x->type)
    {
    case EX_AND:
    case EX_OR:
    case EX_NOT:
      if (!find_words(x->u.op.l) || x->u.op.r && !find_words(x->u.op.r))
	return 0;
      break;
    case EX_MATCH:
      for(i=0; i<num_words; i++)
	if (words[i].is_nonword == x->u.match.is_nonword && !strcmp(words[i].text, x->u.match.word))
	  goto match1;
      if (num_words >= max_words || num_words >= HARD_MAX_WORDS)
	return 0;
      num_words++;
      strcpy(words[i].text, x->u.match.word);
      words[i].is_nonword = x->u.match.is_nonword;
      words[i].firstr = NULL;
    match1:
      w = words + i;
      for(i=0; i<num_word_refs; i++)
	if (wordrefs[i].kw == w && wordrefs[i].classmap == x->u.match.classmap)
	  goto match2;
      if (num_word_refs >= max_words || num_word_refs >= HARD_MAX_WORDS)
	return 0;
      num_word_refs++;
      wordrefs[i].kw = w;
      wordrefs[i].classmap = x->u.match.classmap;
      wordrefs[i].qfactor = 0;
      wordrefs[i].next = words[i].firstr;
      words[i].firstr = &wordrefs[i];
      wordrefs[i].mask = 1 << i;
    match2:
      wordrefs[i].qfactor += x->u.match.quality;
      x->u.match.wd_index = i;
      DBG("`%s' -> %d (%d)", x->u.match.word, i, w - words);
      break;
    case EX_ANY:
      break;
    default:
      die("What do you mean by showing extype=%d?", x->type);
    }
  return 1;
}

static int
solve_words(struct query *q)
{
  uns i;
  int j;

  for(i=0; i<num_words; i++)
    {
      struct known_word *w = words + i;
      j = (w->is_nonword ? lookup_nonword : lookup_word)(w->text, w->refs);
      DBG("Looking up %s (t=%d) -> %d refs", w->text, w->is_nonword, j);
      if (j < 0)
	{
	  prepend_reply(query_rbuf, "-0106 Too many words matching <%s> [max is <%d>]", w->text, max_word_matches);
	  return 0;
	}
      add_reply(cache_rbuf, "W%d\t%s", j, w->text);
      w->nrefs = j;
    }
  return 1;
}

/* Boolean expression evaluation */

static byte bool_map[ 1 << (HARD_MAX_WORDS - 3) ];
static uns bool_bytes;

static void
map_bool(struct expr *x, byte *m)
{
  byte *a, *b;
  ulg *X, *Y, *Z;
  uns c;

  Z = (ulg *) m;
  switch (x->type)
    {
    case EX_AND:
    case EX_OR:
      a = alloca(bool_bytes);
      b = alloca(bool_bytes);
      map_bool(x->u.op.l, a);
      map_bool(x->u.op.r, b);
      X = (ulg *) a;
      Y = (ulg *) b;
      if (x->type == EX_AND)
	for(c=0; c<bool_bytes; c += sizeof(ulg))
	  *Z++ = *X++ & *Y++;
      else
	for(c=0; c<bool_bytes; c += sizeof(ulg))
	  *Z++ = *X++ | *Y++;
      break;
    case EX_NOT:
      a = alloca(bool_bytes);
      map_bool(x->u.op.l, a);
      X = (ulg *) a;
      for(c=0; c<bool_bytes; c += sizeof(ulg))
	*Z++ = ~*X++;
      break;
    case EX_MATCH:
      {
	uns i = x->u.match.wd_index;
	if (i < ulg_bits_log)
	  {
	    uns k = 1 << i;
	    uns t;
	    ulg j = 0;
	    for(t=0; t < 8*sizeof(ulg); t++)
	      if (t & k)
		j |= 1 << t;
	    for(c=0; c<bool_bytes; c += sizeof(ulg))
	      *Z++ = j;
	  }
	else
	  {
	    uns per = 1 << (i - ulg_bits_log);
	    uns t = 2*per*sizeof(ulg);
	    uns i;
	    for(c=0; c<bool_bytes; c += t)
	      {
		for(i=0; i<per; i++)
		  *Z++ = 0;
		for(i=0; i<per; i++)
		  *Z++ = ~0;
	      }
	  }
	break;
      }
    case EX_ANY:
      memset(m, ~0, bool_bytes);
      break;
    default:
      die("Lost in mist and scared of seeing ghost #%d", x->type);
    }
#if 0
  {
    int o;
    for(o=bool_bytes; --o >= 0;)
      printf("%02x", m[o]);
    putchar('\n');
  }
#endif
}

static void
construct_bool(struct expr *x)
{
  if (num_word_refs <= ulg_bits_log)
    bool_bytes = sizeof(ulg);
  else
    bool_bytes = 1 << (num_word_refs - 3);
  DBG("Constructing boolean expression map of size %d", bool_bytes);
  map_bool(x, bool_map);
}

/* Shortcut solution testing */

static int				/* <0=false, >0=true, 1=don't know */
shortcut_solution(struct expr *x)
{
  int l, r;

  switch (x->type)
    {
    case EX_AND:
      l = shortcut_solution(x->u.op.l);
      r = shortcut_solution(x->u.op.r);
      if (l > 0 && r > 0)
	return 1;
      if (l < 0 || r < 0)
	return -1;
      return 0;
    case EX_OR:
      l = shortcut_solution(x->u.op.l);
      r = shortcut_solution(x->u.op.r);
      if (l > 0 || r > 0)
	return 1;
      if (l < 0 && r < 0)
	return -1;
      return 0;
    case EX_NOT:
      l = shortcut_solution(x->u.op.l);
      return -l;
    case EX_MATCH:
      return (wordrefs[x->u.match.wd_index].kw->nrefs) ? 0 : -1;
    case EX_ANY:
      return 1;
    default:
      die("shortcut_solution: black magic (%d)", x->type);
    }
}

/* Reference processing */

static int ref_fd;

struct ref_ref {
  ulg ref;
  int q;
};

static struct ref_ref *ref_heap;

struct ref_src {
  node n;
  ulg pos;
  void *locked_page;
  struct sdbitem *refs;
  uns nrefs, mrefs;
  struct known_word_ref *kwr;
};

static struct ref_src *ref_sources;
static list ref_src_queue;
static uns ref_cnt;
static byte *section_bitmap;

static void
get_ref_srcs(void)
{
  uns i, j, l;
  struct ref_src *r = ref_sources;
  ulg *z;

  init_list(&ref_src_queue);
  ref_cnt = 0;
  for(i=0; i<num_words; i++)
    {
      struct known_word *w = &words[i];
      uns nmat = 0;
      for(j=0; j<w->nrefs; j++)
	{
	  add_tail(&ref_src_queue, &r->n);
	  r->pos = w->refs[j];
	  z = (ulg *) read_data(ref_fd, r->pos, &l);
	  if (l < sizeof(ulg))
	    die("Internal error: Misaligned reference entry!");
	  r->pos += sizeof(ulg);
	  r->locked_page = lock_last_page();
	  r->refs = (struct sdbitem *) (z+1);
	  r->nrefs = (FILE_PAGE_SIZE - (r->pos & (FILE_PAGE_SIZE-1))) / sizeof(struct sdbitem);
	  r->mrefs = *z;
	  if (r->nrefs > r->mrefs)
	    r->nrefs = r->mrefs;
	  nmat += r->mrefs;
	  r->mrefs -= r->nrefs;
	  r->kwr = w->firstr;
	  DBG("REF %d: @%d, %d cached, %d more", ref_cnt, r->pos, r->nrefs, r->mrefs);
	  r++;
	  ref_cnt++;
	}
      add_reply(cache_rbuf, "w%d\t%s", nmat, w->text);
    }
}

static void
clean_ref_srcs(void)
{
  uns i;

  for(i=0; i<ref_cnt; i++)
    unlock_page(ref_sources[i].locked_page);
}

static uns
fill_in_more(struct ref_src *r)
{
  if (!r->mrefs)
    {
      rem_node(&r->n);
      return 0;
    }
  unlock_page(r->locked_page);
  r->pos = (r->pos + FILE_PAGE_SIZE - 1) & ~(FILE_PAGE_SIZE-1);
  r->refs = (struct sdbitem *) read_page(ref_fd, r->pos);
  r->locked_page = lock_last_page();
  r->nrefs = FILE_PAGE_SIZE / sizeof(struct sdbitem);
  if (r->nrefs > r->mrefs)
    r->nrefs = r->mrefs;
  r->mrefs -= r->nrefs;
  DDBG("Loaded %d more for channel %d (%d remains)", r->nrefs, r - ref_sources, r->mrefs);
  return 1;
}

static inline ulg
get_obj_id(void)
{
  struct ref_src *rs, *rn, *rb;
  ulg best;

again:
  best = ~0;
  rb = NULL;
  for(rs = HEAD(ref_src_queue); rn = (void *) rs->n.next; rs = rn)
    {
      if (!rs->nrefs && !fill_in_more(rs))
	continue;
      if (rs->refs->oid < best)
	{
	  best = rs->refs->oid;
	  rb = rs;
	}
    }
  if (rb)
    {
      uns sec = rb->refs->tag;
      if (! (section_bitmap[sec >> 3U] & (1 << (sec & 7))))
	{
	  DDBG("OID %08x: bad sec %d", best, sec);
	  DO_FOR_ALL(rs, ref_src_queue)
	    {
	      rs->refs++;
	      rs->nrefs--;
	    }
	  goto again;
	}
    }
  DDBG("OID %08x", best);
  return best;
}

static inline uns
get_obj_q(ulg obj)
{
  struct ref_src *rs;
  struct known_word_ref *w;
  uns bitmap = 0;
  uns i;
  int bq, allq, q;
  ulg classmask;

  for(i=0; i<num_word_refs; i++)
    wordrefs[i].qtemp = 0;

  DO_FOR_ALL(rs, ref_src_queue)
    {
      if (rs->refs->oid != obj)
	continue;
      classmask = 1 << WORDTYPE(rs->refs->flags);
      q = rs->refs->flags & WQ_MASK;
      bq = -1;
      for(w = rs->kwr; w; w = w->next)
	{
	  if (!(w->classmap & classmask))
	    continue;
	  if (bq < (int) w->qfactor)
	    bq = w->qfactor;
	}
      if (bq >= 0)
	{
	  bq = (bq - 0x10) * q + found_word_bonus;
	  bitmap |= rs->kwr->mask;
	  if (bq >= rs->kwr->qtemp)
	    rs->kwr->qtemp = bq;
	}
      rs->refs++;
      rs->nrefs--;
    }
  if (! (bool_map[bitmap >> 3U] & (1 << (bitmap & 7))))
    {
      DDBG("Bitmap %x, rejected", bitmap);
      return ~0;
    }

  allq = 0;
  for(i=0; i<num_word_refs; i++)
    allq += wordrefs[i].qtemp;
  DDBG("Bitmap %x, q=%d, accepted", bitmap, allq);
  return (allq >= 0) ? allq : 0;
}

static inline void
heap_remove(uns n)
{
  struct ref_ref *h = ref_heap - 1;	/* We will use indices starting with 1 */
  uns p, z, zz;
  struct ref_ref t;

  h[1] = h[n];
  p = 1;
  for(;;)
    {
      z = 2*p;
      if (z >= n)
	break;
      if ((zz = z+1) >= n)
	zz = z;
      if (h[z].q < h[p].q || h[zz].q < h[p].q)
	{
	  if (h[z].q > h[zz].q)
	    z = zz;
	  t = h[z]; h[z] = h[p]; h[p] = t;
	  p = z;
	}
      else
	break;
    }
}

static inline void
heap_insert(ulg oid, uns q, uns n)
{
  struct ref_ref *h = ref_heap - 1;	/* We will use indices starting with 1 */
  uns p;

  n++;					/* Position of the new element */
  h[n].ref = oid;
  h[n].q = q;
  while ((p = n/2) && (h[p].q > h[n].q))
    {
      struct ref_ref t;
      t = h[p]; h[p] = h[n]; h[n] = t;
      n = p;
    }
}

static uns
fill_heap(void)
{
  uns heap_count = 0;
  ulg obj;
  uns q;
  uns total_objs = 0;

  for(;;)
    {
      obj = get_obj_id();
      if (obj == ~0)			/* END */
	break;
      q = get_obj_q(obj);
      if (q == ~0)
	continue;
      if (heap_count >= num_matches)
	heap_remove(heap_count--);
      heap_insert(obj, q, heap_count++);
      total_objs++;
    }
  add_reply(cache_rbuf, "T%d", total_objs);
  return heap_count;
}

static uns
do_refs(struct query *q, ulg *oids, ulg *qs)
{
  uns heap_count, i;

  section_bitmap = q->sections;
  get_ref_srcs();
  if (i = heap_count = fill_heap())
    while (i)
      {
	--i;
	oids[i] = ref_heap[0].ref;
	qs[i] = ref_heap[0].q;
	DDBG("Ref %d: %08x q=%d", i, oids[i], qs[i]);
	heap_remove(i+1);
      }
  clean_ref_srcs();
  return heap_count;
}

void
ref_init(void)
{
  ref_fd = open_file(ref_list_name);
  ref_heap = xmalloc(sizeof(struct ref_ref) * num_matches);
  ref_sources = xmalloc(sizeof(struct ref_src) * max_words * max_word_matches);
}

/* If a cache lookup fails, we must do the hard work... */

static int
process_query(struct query *q, struct cache_item *c)
{
  num_words = 0;
  num_word_refs = 0;
  DBG("Unifying words");
  if (!find_words(q->expr))
    {
      prepend_reply(query_rbuf, "-0103 Too many words");
      return 0;
    }
  construct_bool(q->expr);
  if (bool_map[0] & 1)			/* Matched by document with none of given words? */
    {
      prepend_reply(query_rbuf, "-0104 Negative queries not supported");
      return 0;
    }
  q->map_msec = get_timer();
  DBG("Searching for words");
  if (!solve_words(q))
    return 0;
  q->words_msec = get_timer();
  DBG("Checking no solution shortcuts");
  if (shortcut_solution(q->expr) < 0)
    {
      add_reply(cache_rbuf, ".No solution. [shortcut]");
      add_reply(cache_rbuf, "T0");
      c->len = 0;
      return 1;
    }
  DBG("Processing references");
  c->len = do_refs(q, c->oids, c->qs);
  DBG("Found %d objects", c->len);
  return 1;
}

/* Dumping of objects */

static void
show_objects(struct query *q, struct cache_item *c, uns start, uns count)
{
  while (count--)
    {
      add_reply(query_rbuf, "Q%d", c->qs[start]);
      show_object(q, c->oids[start]);
      flush_reply_buf(q, query_rbuf);
      reply_string(q, "\n", 1);
      start++;
    }
}
