/*
 *	Sherlock Search Engine -- Searching for Words
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "sherlock.h"

#include <lib/searchdb.h>
#include <lib/string.h>
#include <lib/wordcmp.h>

static int tree_fd, wlist_fd;
struct tree_page tree_root;

/* BTREE */

#define MAX_DEPTH 4

static uns root_pointers[KEYS_PER_PAGE + 1], root_sizes[KEYS_PER_PAGE + 1],
           rel_pointers[KEYS_PER_PAGE + 1][MAX_DEPTH], depths[KEYS_PER_PAGE + 1];

static void
btree_init(void)
{
  struct tree_page *root;
  uns keys, skip, i, j, w, k, p, ww, pos;
  uns width[16], depth;

  root = (struct tree_page *) read_data(tree_fd, 0, &i);
  memcpy(&tree_root, root, sizeof(tree_root));
  keys = root->aux;			/* See search/build/wtree.c for an explanation of what we're doing here */
  skip = keys / (KEYS_PER_PAGE + 1);
  i = 0;
  pos = TREE_PAGE_SIZE;
  for(j=0; j<=KEYS_PER_PAGE; j++)
    {
      w = skip;
      if (i + w >= keys)
	w = keys - i;			/* Subtree: <w> keys starting with the <i>-th */

      width[0] = 0;			/* Calculate subtree width */
      depth = 0;
      while (width[depth] < w)
	{
	  width[depth+1] = width[depth] * (KEYS_PER_PAGE+1) + KEYS_PER_PAGE;
	  depth++;
	}
      depths[j] = depth;
      if (depth > MAX_DEPTH)
	die("Tree too high to climb on");
      p = 0;				/* Total pages per subtree */
      for(k=0; k<depth; k++)
	{
	  rel_pointers[j][k] = p;
	  ww = width[depth - k - 1] * (KEYS_PER_PAGE + 1) + KEYS_PER_PAGE + 1;
	  p += (w + ww - 2) / ww;
	}
      root_sizes[j] = p;
      root_pointers[j] = pos;
      DDBG("subtree=%d first=%d keys=%d depth=%d pages=%d pkeys=%d pos=%d len=%d",
	  j, i, w, depth, p, p*KEYS_PER_PAGE, pos, p * TREE_PAGE_SIZE);
      pos += p * TREE_PAGE_SIZE;
      i += skip + 1;
    }
  DBG("Total tree file size should be %d", pos);
  if (lseek(tree_fd, 0, SEEK_END) != pos)
    die("Unexpected inconsistency in the btree: length should be %d, but it isn't!", pos);
}

static int
lookup_bt_key(struct tree_page *p, byte *key)
{
  int l = 0;
  int r = KEYS_PER_PAGE - 1;
  int m = 0;
  int z;

  while (l <= r)
    {
      m = (l + r) / 2;
      DDBG("/%-8.8s/%-8.8s/", p->keys[m].sig, key);
      z = memcmp(key, p->keys[m].sig, 8);
      DDBG("l=%d m=%d r=%d z=%d", l,m,r,z);
      if (!z)
	return -m-1;
      if (z < 0)
	r = m - 1;
      else
	l = m + 1;
    }
  z = memcmp(key, p->keys[m].sig, 8);
  return (z > 0) ? (m + 1) : m;
}

static ulg
find_sig(byte *sig)
{
  struct tree_page *r = &tree_root;
  int i, z;
  uns base, pag, npag, dep;

  i = z = lookup_bt_key(r, sig);
  if (i < 0)
    goto found;
  base = root_pointers[z];
  npag = root_sizes[z];
  pag = 0;
  dep = 0;
  DDBG("Going to subtree %d, base=%d, npag=%d", z, base, npag);
  while (pag < npag)
    {
      r = (struct tree_page *) read_data(tree_fd, TREE_PAGE_SIZE*pag + base, &i);
      i = lookup_bt_key(r, sig);
      if (i < 0)
	goto found;
      dep++;
      if (dep > depths[z])
	{
	  DDBG("Too deep. [DG]iving up.");
	  return ~0;
	}
      pag = rel_pointers[z][dep] + i;
      DDBG("Going to page %d", pag);
    }
  DDBG("Fell off the tree");
  return ~0;

found:
  i = -i-1;
  DDBG("Found at position %d, sigpos=%d", i, r->keys[i].ptr);
  return r->keys[i].ptr;
}

/* Filtering of words */

static inline int
is_accented(byte *w)
{
  while (*w)
    if (_c_order[*w++])
      return 1;
  return 0;
}

static inline int
match_detail(byte *buf, uns blen, byte *w, byte *unacc, uns wlen)
{
  if (blen != wlen)
    return 0;
  if (is_accented(buf))			/* Very magic -- here are the rules for accented/non-accented comparison */
    {
      if (unacc)
	{
	  while (*unacc)
	    if (*unacc++ != _c_collate[*buf++])
	      return 0;
	  return 1;
	}
      return !strcmp(buf, w);
    }
  else
    {
      if (unacc)
	return !strcmp(buf, unacc);
      else
	return !strcmp(buf, w);
    }
}

static int
lookup_detail(byte *w, byte *unacc, uns len, ulg *x, ulg pos)
{
  uns cnt = 0;
  byte buf[MAX_WORD_LEN];
  byte *b = buf;
  byte *r, *e;
  uns s = 0;
  uns l, wl = 0;
  ulg ref;

  for(;;)
    {
      r = read_data(wlist_fd, pos, &l);
      pos += l;
      e = r + l;
      while (r < e)
	{
	  if (s)
	    {
	      *b++ = *r++;
	      if (!--s)
		{
		  DDBG("/%s/ %d", buf, ref);
		  if (match_detail(buf, wl, w, unacc, len))
		    {
		      if (cnt++ >= max_word_matches)
			{
			  DBG("Too many matches");
			  return -1;
			}
		      *x++ = ref;
		      DDBG("MATCH");
		    }
		  b = buf;
		}
	    }
	  else if (*r)
	    *b++ = *r++;
	  else
	    {
	      *b = *r++;
	      wl = b - buf;
	      if (!wl)
		return cnt;
	      b = (byte *) &ref;
	      s = sizeof(ref);
	    }
	}
    }
}

/* Main word lookup function */

int
lookup_word(byte *w, ulg *x)
{
  int wild = 0;
  int accent = 0;
  byte *a, *b, *unacc;
  byte signature[8];
  ulg sigpos;
  uns wlen;

  for(a=w; *a; a++)
    if (*a == '?' || *a == '*')
      wild = 1;
    else if (Cspace(*a))
      {
	DBG("Shortcut: spaces -> not a word");
	return 0;
      }
    else if (_c_order[*a])
      accent++;
  wlen = a - w;

  if (wild)
    {
      log(L_ERROR "Wildcard search requested, but not supported.");
      return 0;
    }

  extract_sig(w, signature);
  sigpos = find_sig(signature);
  if (sigpos == ~0)
    return 0;

  if (accent)
    {
      b = unacc = alloca(wlen + 1);
      for(a=w; *b++ = _c_collate[*a++];)
	;
      DBG("Unaccented variant is `%s'", unacc);
    }
  else
    unacc = NULL;

  return lookup_detail(w, unacc, wlen, x, sigpos);
}

void
word_init(void)
{
  tree_fd = open_file(sig_tree_name);
  wlist_fd = open_file(word_list_name);
  btree_init();
}

byte *
normalize_word(struct mempool *p, byte *x)
{
  byte *k = x;
  while (*k)
    {
      *k = Cupcase(*k);
      k++;
    }
  return x;
}
